import datetime as dt
import logging
from typing import List, Dict

import pandas as pd

import requests

base_url = 'https://statsapi.mlb.com/api/v1/'

urls = {
    'schedule_by_day': base_url + "schedule/games/?sportId=1&date={run_date}",
    'line_ups': base_url + 'schedule?gamePk={game_pk}&language=en&hydrate=lineups,probablePitcher(note)',
    'game': base_url + "game/{game_pk}/boxscore?timecode={run_date}",
    'schedule': base_url + "schedule/?eventTypes=primary,secondary&sportId=1&startDate={start_date}&endDate={end_date}"
}

headers = {
        "Content-Type": "application/json"
}

logger = logging.getLogger(__name__)


class MLBStats(object):
    """
    Class for MLBStats

    """
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.logger.info('MLBStats Initialized')

    def get_game_lineups(self, game_pk) -> pd.DataFrame:
        """
        Get Lineups for Today's Games

        :return:
        """
        self.logger.info('Getting Lineups|{}'.format(urls['line_ups'].format(game_pk=game_pk)))
        lineups = []

        # Send a GET request to the API endpoint
        response = requests.get(urls['line_ups'].format(game_pk=game_pk), headers=headers)

        # Check if the request was successful
        if response.status_code == 200:
            # Get the JSON response
            data = response.json()

            # Check if the game data is available
            if "dates" in data and len(data["dates"]) > 0 and "games" in data["dates"][0] and len(
                    data["dates"][0]["games"]) > 0:
                game = data["dates"][0]["games"][0]

                # Get the projected lineups
                for team in ['home', 'away']:
                    try:
                        lineup = game["lineups"][f"{team}Players"]
                    except KeyError:
                        print(game)
                        logger.info(f"No projected lineup for {team} team")
                        logger.info(game)
                        continue
                    for i, player_info in enumerate(lineup):
                        player = {
                                "date":          game["gameDate"],
                                "player_id":     player_info["id"],
                                "player_name":   player_info["fullName"],
                                "position":      player_info["primaryPosition"]["abbreviation"],
                                "batting_order": i + 1,
                                "team":          {
                                        "team_id":   data["dates"][0]["games"][0]["teams"][team]["team"]["id"],
                                        "team_name": data["dates"][0]["games"][0]["teams"][team]["team"]["name"],
                                }
                        }
                        lineups.append(player)
                return pd.DataFrame(lineups)
            else:

                raise Exception(f"Error retrieving the projected lineup for Game ID: {game_pk}")

        else:
            raise Exception(f"Error retrieving the projected lineup for Game ID: {game_pk}")

    @staticmethod
    def _process_game(game_data):
        """
        Process game data

        :param game_data:
        :return:
        """
        pass

    def get_lineups_by_day(self, run_date: dt.datetime) -> pd.DataFrame:
        """
        Get Schedule by Day

        :param run_date:
        :return:
        """
        self.logger.info('Getting Lineups|{}'.format(urls['schedule_by_day'].format(run_date=run_date)))
        games = []

        # Send a GET request to the API endpoint
        response = requests.get(urls['schedule_by_day'].format(run_date=run_date.strftime("%Y-%m-%d")), headers=headers)

        # Check if the request was successful
        if response.status_code == 200:
            # Get the JSON response
            data = response.json()

            # Check if the game data is available
            for date in data["dates"]:
                for game in date["games"]:
                    return self.get_game_lineups(game["gamePk"])

    def get_game_boxscore(self, game_pk) -> pd.DataFrame:
        """
        Get Game Logs

        :param game_pk:
        :return:
        """
        self.logger.info('Getting Game Logs|{}'.format(urls['game'].format(run_date=game_pk)))

        # Send a GET request to the API endpoint
        response = requests.get(urls['game'].format(run_date=game_pk), headers=headers)

        # Check if the request was successful
        if response.status_code == 200:
            # Get the JSON response
            data = response.json()

            # Check if the game data is available
            if "dates" in data and len(data["dates"]) > 0 and "games" in data["dates"][0] and len(
                    data["dates"][0]["games"]) > 0:
                game = data["dates"][0]["games"][0]

                return game
            else:
                raise Exception(f"Error retrieving the game data for Game ID: {game_pk}")
        else:
            raise Exception(f"Error retrieving the game data for Game ID: {game_pk}")

    def get_schedule(self, start_date: dt.datetime, end_date: dt.datetime) -> pd.DataFrame:
        """
        Get Schedule

        :return:
        """
        self.logger.info('Getting Schedule|{}'.format(urls['schedule']))
        games = []

        # Send a GET request to the API endpoint
        response = requests.get(urls['schedule'].format(
            start_date=start_date.strftime("%Y-%m-%d"),
            end_date=end_date.strftime("%Y-%m-%d"),
            headers=headers))

        # Check if the request was successful
        if response.status_code == 200:
            # Get the JSON response
            data = response.json()

            # Check if the game data is available
            for date in data["dates"]:
                for game in date["games"]:
                    games.append(game)

            return pd.DataFrame(games)
        else:
            raise Exception(f"Error retrieving the schedule")



"""
Retrosheet Data Notice:

Recipients of Retrosheet data are free to make any desired use of
the information, including (but not limited to) selling it,
giving it away, or producing a commercial product based upon the
data.  Retrosheet has one requirement for any such transfer of
data or product development, which is that the following
statement must appear prominently:

     The information used here was obtained free of
     charge from and is copyrighted by Retrosheet.  Interested
     parties may contact Retrosheet at "www.retrosheet.org".

Retrosheet makes no guarantees of accuracy for the information
that is supplied. Much effort is expended to make our website
as correct as possible, but Retrosheet shall not be held
responsible for any consequences arising from the use the
material presented here. All information is subject to corrections
as additional data are received. We are grateful to anyone who
discovers discrepancies and we appreciate learning of the details.
"""
import pandas as pd
from plyball.utils import get_zip_file
from datetime import datetime


class RetroSheet(object):
    _game_log_columns = ['date', 'game_num', 'day_of_week', 'visiting_team',
                       'visiting_team_league', 'visiting_game_num', 'home_team',
                       'home_team_league', 'home_team_game_num', 'visiting_score',
                       'home_score', 'num_outs', 'day_night', 'completion_info',
                       'forfeit_info', 'protest_info', 'park_id', 'attendance',
                       'time_of_game_minutes', 'visiting_line_score',
                       'home_line_score', 'visiting_abs', 'visiting_hits',
                       'visiting_doubles', 'visiting_triples', 'visiting_homeruns',
                       'visiting_rbi', 'visiting_sac_hits', 'visiting_sac_flies',
                       'visiting_hbp', 'visiting_bb', 'visiting_iw', 'visiting_k',
                       'visiting_sb', 'visiting_cs', 'visiting_gdp', 'visiting_ci',
                         'visiting_lob', 'visiting_pitchers_used',
                         'visiting_individual_er', 'visiting_er', 'visiting__wp',
                         'visiting_balks', 'visiting_po', 'visiting_assists',
                         'visiting_errors', 'visiting_pb', 'visiting_dp',
                         'visiting_tp', 'home_abs', 'home_hits', 'home_doubles',
                         'home_triples', 'home_homeruns', 'home_rbi',
                         'home_sac_hits', 'home_sac_flies', 'home_hbp', 'home_bb',
                         'home_iw', 'home_k', 'home_sb', 'home_cs', 'home_gdp',
                         'home_ci', 'home_lob', 'home_pitchers_used',
                         'home_individual_er', 'home_er', 'home_wp', 'home_balks',
                         'home_po', 'home_assists', 'home_errors', 'home_pb',
                         'home_dp', 'home_tp', 'ump_home_id', 'ump_home_name',
                         'ump_first_id', 'ump_first_name', 'ump_second_id',
                         'ump_second_name', 'ump_third_id', 'ump_third_name',
                         'ump_lf_id', 'ump_lf_name', 'ump_rf_id', 'ump_rf_name',
                         'visiting_manager_id', 'visiting_manager_name',
                         'home_manager_id', 'home_manager_name',
                         'winning_pitcher_id', 'winning_pitcher_name',
                         'losing_pitcher_id', 'losing_pitcher_name',
                         'save_pitcher_id', 'save_pitcher_name',
                         'game_winning_rbi_id', 'game_winning_rbi_name',
                         'visiting_starting_pitcher_id',
                         'visiting_starting_pitcher_name',
                         'home_starting_pitcher_id', 'home_starting_pitcher_name',
                         'visiting_1_id', 'visiting_1_name', 'visiting_1_pos',
                         'visiting_2_id', 'visiting_2_name', 'visiting_2_pos',
                         'visiting_2_id.1', 'visiting_3_name', 'visiting_3_pos',
                         'visiting_4_id', 'visiting_4_name', 'visiting_4_pos',
                         'visiting_5_id', 'visiting_5_name', 'visiting_5_pos',
                         'visiting_6_id', 'visiting_6_name', 'visiting_6_pos',
                         'visiting_7_id', 'visiting_7_name', 'visiting_7_pos',
                         'visiting_8_id', 'visiting_8_name', 'visiting_8_pos',
                         'visiting_9_id', 'visiting_9_name', 'visiting_9_pos',
                         'home_1_id', 'home_1_name', 'home_1_pos', 'home_2_id',
                         'home_2_name', 'home_2_pos', 'home_3_id', 'home_3_name',
                         'home_3_pos', 'home_4_id', 'home_4_name', 'home_4_pos',
                         'home_5_id', 'home_5_name', 'home_5_pos', 'home_6_id',
                         'home_6_name', 'home_6_pos', 'home_7_id', 'home_7_name',
                         'home_7_pos', 'home_8_id', 'home_8_name', 'home_8_pos',
                         'home_9_id', 'home_9_name', 'home_9_pos', 'misc',
                         'acquisition_info']
    _urls = {'game_log_url': 'http://www.retrosheet.org/gamelogs/gl{}.zip',
            'world_series_url': 'http://www.retrosheet.org/gamelogs/glws.zip',
            'all_star_url': 'http://www.retrosheet.org/gamelogs/glas.zip',
            'wild_card_url': 'http://www.retrosheet.org/gamelogs/glwc.zip',
            'division_series_url': 'http://www.retrosheet.org/gamelogs/gldv.zip',
            'lcs_url': 'http://www.retrosheet.org/gamelogs/gllc.zip',
             }

    def season_game_logs(self, season):
        """
        Pull Retrosheet game logs for a given season
        :param season: Game Logs for Season

        :return: DataFrame
        """
        # validate input
        max_year = int(datetime.now().year) - 1
        if season > max_year or season < 1871:
            raise ValueError('Season must be between 1871 and {}'.format(max_year))
        file_name = 'GL{}.TXT'.format(season)
        z = get_zip_file(self.gamelog_url.format(season))
        data = pd.read_csv(z.open(file_name), header=None, sep=',', quotechar='"')
        data.columns = self._game_log_columns
        return data

    @property
    def world_series_logs(self):
        """
        Pull Retrosheet World Series Game Logs
        :return: DataFrame
        """
        file_name = 'GLWS.TXT'
        z = get_zip_file(self._urls['world_series_url'])
        data = pd.read_csv(z.open(file_name), header=None, sep=',', quotechar='"')
        data.columns = self._game_log_columns
        return data

    @property
    def all_star_game_logs(self):
        """
        Pull Retrosheet All Star Game Logs
        :return: DataFrame
        """
        file_name = 'GLAS.TXT'
        z = get_zip_file(self._urls['all_star_url'])
        data = pd.read_csv(z.open(file_name), header=None, sep=',', quotechar='"')
        data.columns = self._game_log_columns
        return data

    @property
    def wild_card_logs(self):
        """
        Pull Retrosheet Wild Card Game Logs
        :return: DataFrame
        """
        file_name = 'GLWC.TXT'
        z = get_zip_file(self._urls['wild_card_ur'])
        data = pd.read_csv(z.open(file_name), header=None, sep=',', quotechar='"')
        data.columns = self._game_log_columns
        return data

    @property
    def division_series_logs(self):
        """
        Pull Retrosheet Division Series Game Logs
        :return: DataFrame
        """
        file_name = 'GLDV.TXT'
        z = get_zip_file(self._urls['division_series_url'])
        data = pd.read_csv(z.open(file_name), header=None, sep=',', quotechar='"')
        data.columns = self._game_log_columns
        return data

    @property
    def lcs_logs(self):
        """
        Pull Retrosheet LCS Game Logs
        :return: DataFrame
        """
        file_name = 'GLLC.TXT'
        z = get_zip_file(self._urls['lcs_url'])
        data = pd.read_csv(z.open(file_name), header=None, sep=',', quotechar='"')
        data.columns = self._game_log_columns
        return data



import requests
import io
import zipfile
from datetime import datetime, timedelta
import pandas as pd

# dictionary containing team abbreviations and their first_name year in existence
first_season_map = {'ALT': 1884, 'ANA': 1997, 'ARI': 1998, 'ATH': 1871,
                    'ATL': 1966, 'BAL': 1872, 'BLA': 1901, 'BLN': 1892,
                    'BLU': 1884, 'BOS': 1871, 'BRA': 1872, 'BRG': 1890,
                    'BRO': 1884, 'BSN': 1876, 'BTT': 1914, 'BUF': 1879,
                    'BWW': 1890, 'CAL': 1965, 'CEN': 1875, 'CHC': 1876,
                    'CHI': 1871, 'CHW': 1901, 'CIN': 1876, 'CKK': 1891,
                    'CLE': 1871, 'CLV': 1879, 'COL': 1883, 'COR': 1884,
                    'CPI': 1884, 'DET': 1901, 'DTN': 1881, 'ECK': 1872,
                    'FLA': 1993, 'HAR': 1874, 'HOU': 1962, 'IND': 1878,
                    'KCA': 1955, 'KCC': 1884, 'KCN': 1886, 'KCP': 1914,
                    'KCR': 1969, 'KEK': 1871, 'LAA': 1961, 'LAD': 1958,
                    'LOU': 1876, 'MAN': 1872, 'MAR': 1873, 'MIA': 2012,
                    'MIL': 1884, 'MIN': 1961, 'MLA': 1901, 'MLG': 1878,
                    'MLN': 1953, 'MON': 1969, 'NAT': 1872, 'NEW': 1915,
                    'NHV': 1875, 'NYG': 1883, 'NYI': 1890, 'NYM': 1962,
                    'NYP': 1883, 'NYU': 1871, 'NYY': 1903, 'OAK': 1968,
                    'OLY': 1871, 'PBB': 1890, 'PBS': 1914, 'PHA': 1882,
                    'PHI': 1873, 'PHK': 1884, 'PHQ': 1890, 'PIT': 1882,
                    'PRO': 1878, 'RES': 1873, 'RIC': 1884, 'ROC': 1890,
                    'ROK': 1871, 'SDP': 1969, 'SEA': 1977, 'SEP': 1969,
                    'SFG': 1958, 'SLB': 1902, 'SLM': 1884, 'SLR': 1875,
                    'STL': 1875, 'STP': 1884, 'SYR': 1879, 'TBD': 1998,
                    'TBR': 2008, 'TEX': 1972, 'TOL': 1884, 'TOR': 1977,
                    'TRO': 1871, 'WAS': 1873, 'WES': 1875, 'WHS': 1884,
                    'WIL': 1884, 'WOR': 1880, 'WSA': 1961, 'WSH': 1901,
                    'WSN': 2005}


def validate_datestring(date_text):
    try:
        datetime.strptime(date_text, '%Y-%m-%d')
    except ValueError:
        raise ValueError("Incorrect data format, should be YYYY-MM-DD")


def sanitize_input(start_dt, end_dt, player_id):
    # error if no player ID provided
    if player_id is None:
        raise ValueError("Player ID is required. If you need to find a player's id, try plyball.playerid_"
                         "lookup(last_name, first_name) and use their key_mlbam. If you want statcast data for all "
                         "players, try the statcast() function.")
    # this id should be a string to place inside a _url
    player_id = str(player_id)
    # if no dates are supplied, assume they want yesterday's data
    # send a warning in case they wanted to specify
    if start_dt is None and end_dt is None:
        today = datetime.today()
        start_dt = (today - timedelta(1)).strftime("%Y-%m-%d")
        end_dt = today
        print("Warning: no date range supplied. Returning yesterday's Statcast data. For a different date range, try "
              "get_statcast(start_dt, end_dt).")
    # if only one date is supplied, assume they only want that day's player_type
    # query in this case is from date 1 to date 1
    if start_dt is None:
        start_dt = end_dt
    if end_dt is None:
        end_dt = start_dt
    # now that both dates are not None, make sure they are valid date strings
    validate_datestring(start_dt)
    validate_datestring(end_dt)
    return start_dt, end_dt, player_id


def split_request(start_dt, end_dt, player_id, url):
    """
    Splits Statcast queries to avoid request timeouts
    """
    results = []
    player_id = str(player_id)
    print('Gathering Player Data')
    while start_dt < end_dt:
        remaining = end_dt - start_dt
        delta = min(remaining, timedelta(days=2190))
        next_dt = start_dt + delta
        start_str = start_dt.strftime('%Y-%m-%d')
        end_str = next_dt.strftime('%Y-%m-%d')
        data = requests.get(url.format(start_str, end_str, player_id))
        df = pd.read_csv(io.StringIO(data.text))
        results.append(df)
        start_dt = next_dt + timedelta(days=1)
    return pd.concat(results)


def get_zip_file(url):
    """
    Get zip file from provided URL
    """
    with requests.get(url, stream=True) as f:
        z = zipfile.ZipFile(io.BytesIO(f.content))
    return z



from typing import Dict, List, Union

import numpy as np
import pandas as pd
import requests
import structlog
from bs4 import BeautifulSoup
from pandas import DataFrame


class Ottoneu(object):
    """
    A class for scrapping data from an Ottoneu pipeline league.

    """
    logger = structlog.get_logger("Ottoneu")

    def __init__(self, league_id: int):
        """
        Initialize Ottoneu with the ID of the League you are a part of.

        :param league_id:
        :type league_id: int
        """
        self.league_id = league_id
        self.ottoneu_base_url = 'https://ottoneu.fangraphs.com/{}'.format(league_id)

    @staticmethod
    def _process_player_page(soup: BeautifulSoup, stat_type: str, league: str = 'MLB') -> DataFrame:
        """
        Process a Player's page to extract player_type using Beautiful Soup

        :param soup: BS4 Website
        :param stat_type: 'Pitching' or 'Batting'
        :param league: 'MLB' or 'MILB'
        :type league: str
        :return: DataFrame
        """
        stats = pd.DataFrame()
        league_stats = soup.find_all("h3", text="{} Stats".format(league))
        if league_stats:
            if soup.find_all("h3", text="{} Stats".format(league))[0].parent.parent.find_all("h3", text=stat_type):
                stats = pd.read_html(str(soup.find_all("h3", text="{} Stats".format(league))[0].
                                         parent.parent.find_all("h3", text=stat_type)[0].
                                         parent.parent.find_all('table')))[0]
                stats['position_type'] = stat_type
                stats['league'] = "{} Stats".format(league)
        return stats

    def players(self, positions: List[str] = None) -> Dict[str, Union[DataFrame,
                                                                      Dict[str, DataFrame],
                                                                      Dict[str, DataFrame]]]:
        """
        Get DataFrame of players for Players listed.

        :param positions: ['All', 'C', '1B', '2B', '3B', 'SS', 'OF', 'SP', 'RP', 'UTIL']
        :type positions: List[str]
        :return: Dictionary consisting of various DataFrames of Players (and their Stats).
        """

        if positions is None:
            positions = ['C', '1B', '2B', '3B', 'SS', 'OF', 'SP', 'RP']

        headers = {
            'authority': 'ottoneu.fangraphs.com',
            'Content-Type': 'application/x-www-form-urlencoded',
            'accept': '*/*',
            'origin': 'https://ottoneu.fangraphs.com',
            'x-requested-with': 'XMLHttpRequest',
            'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.117 Safari/537.36',
            'content-player_type': 'application/x-www-form-urlencoded',
            'sec-fetch-site': 'same-origin',
            'sec-fetch-mode': 'cors',
            'referer': '{}/search'.format(self.ottoneu_base_url),
            'accept-encoding': 'gzip, deflate, br',
            'accept-language': 'en-US,en;q=0.9'
        }

        json_data = {
            'data': {
                'txtSearch': '',
                'selPos': positions,
                'playerLevel': 'majors',
                'chkFAOnly': '',
                'searchFilter': '',
                'searchComparison': '',
                'searchQualification': '',
            }
        }

        data = ''

        for wrapper, parameters in json_data.items():
            for param, value in parameters.items():
                if type(value) == list:
                    for selection in value:
                        data = data + '{}%5B{}%5D%5B%5D={}'.format(wrapper, param, selection)
                        data = data + '&'
                else:
                    data = data + '{}%5B{}%5D={}'.format(wrapper, param, value)
                data = data + '&'

        self.logger.info("{}/ajax/search".format(self.ottoneu_base_url))
        self.logger.info(data[:-1])

        a = requests.post("{}/ajax/search".format(self.ottoneu_base_url),
                          data=data,
                          headers=headers)

        _json = a.json()
        self.logger.info(f'JSON: {_json}')
        batter_info = []
        batter_stats = []
        for batter in _json['batterResults']:
            batter_info.append({k: v for k, v in batter.items() if k != 'Stats'})
            try:
                __stat_dict = {k: v for k, v in batter['Stats']['batting'].items()}
            except KeyError as e:
                self.logger.info(f"Batter: {batter['PlayerID']}|{e}")
                __stat_dict = {}
            except TypeError as e:
                self.logger.info(f"Batter: {batter['PlayerID']}|{e}")
                __stat_dict = {}
            __stat_dict['PlayerID'] = batter['PlayerID']
            batter_stats.append(__stat_dict)

        pitcher_info = []
        pitcher_stats = []
        for pitcher in _json['pitcherResults']:
            pitcher_info.append({k: v for k, v in pitcher.items() if k != 'Stats'})
            try:
                __stat_dict = {k: v for k, v in pitcher['Stats']['pitching'].items()}
            except KeyError as e:
                self.logger.info(f"Pitcher: {pitcher['PlayerID']}|{e}")
                __stat_dict = {}
            except TypeError as e:
                self.logger.info(f"Pitcher: {pitcher['PlayerID']}|{e}")
                __stat_dict = {}

            __stat_dict['PlayerID'] = pitcher['PlayerID']
            pitcher_stats.append(__stat_dict)

        df_batter_info = pd.DataFrame(batter_info)
        df_pitcher_info = pd.DataFrame(pitcher_info)

        df_batter_stat = pd.DataFrame(batter_stats)
        df_pitcher_stat = pd.DataFrame(pitcher_stats)

        # concat batter and pitcher info and stats rows into one dataframe
        df_info = pd.concat([df_batter_info, df_pitcher_info], sort=False)
        df_stat = pd.concat([df_batter_stat, df_pitcher_stat], sort=False)
        # df_info = df_batter_info.concat([df_pitcher_info], sort=False)
        # df_stat = df_batter_stat.append(df_pitcher_stat, sort=False)

        return {
            'info': df_info,
            'stat': df_stat,
            'batter': {
                'info': df_batter_info,
                'stat': df_batter_stat
            },
            'pitcher': {
                'info': df_pitcher_info,
                'stat': df_pitcher_stat
            }
        }

    def minor_players(self, positions: List[str] = None) -> Dict[str, Union[DataFrame,
    Dict[str, DataFrame],
    Dict[str, DataFrame]]]:
        """
        Get DataFrame of players for Players listed.

        :param positions: ['All', 'C', '1B', '2B', '3B', 'SS', 'OF', 'SP', 'RP', 'UTIL']
        :type positions: List[str]
        :return: Dictionary consisting of various DataFrames of Players (and their Stats).
        """

        if positions is None:
            positions = ['C', '1B', '2B', '3B', 'SS', 'OF', 'SP', 'RP']

        headers = {
            'authority': 'ottoneu.fangraphs.com',
            'Content-Type': 'application/x-www-form-urlencoded',
            'accept': '*/*',
            'origin': 'https://ottoneu.fangraphs.com',
            'x-requested-with': 'XMLHttpRequest',
            'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.117 Safari/537.36',
            'content-player_type': 'application/x-www-form-urlencoded',
            'sec-fetch-site': 'same-origin',
            'sec-fetch-mode': 'cors',
            'referer': '{}/search'.format(self.ottoneu_base_url),
            'accept-encoding': 'gzip, deflate, br',
            'accept-language': 'en-US,en;q=0.9'
        }

        json_data = {
            'data': {
                'txtSearch': '',
                'selPos': positions,
                'playerLevel': 'minors',
                'chkFAOnly': '',
                'searchFilter': '',
                'searchComparison': '',
                'searchQualification': '',
            }
        }

        data = ''

        for wrapper, parameters in json_data.items():
            for param, value in parameters.items():
                if type(value) == list:
                    for selection in value:
                        data = data + '{}%5B{}%5D%5B%5D={}'.format(wrapper, param, selection)
                        data = data + '&'
                else:
                    data = data + '{}%5B{}%5D={}'.format(wrapper, param, value)
                data = data + '&'

        self.logger.info("{}/ajax/search".format(self.ottoneu_base_url))
        self.logger.info(data[:-1])

        a = requests.post("{}/ajax/search".format(self.ottoneu_base_url),
                          data=data,
                          headers=headers)

        _json = a.json()
        self.logger.info(f'JSON: {_json}')
        batter_info = []
        batter_stats = []
        for batter in _json['batterResults']:
            batter_info.append({k: v for k, v in batter.items() if k != 'Stats'})
            try:
                __stat_dict = {k: v for k, v in batter['Stats']['batting'].items()}
            except KeyError as e:
                self.logger.info(f"Batter: {batter['PlayerID']}|{e}")
                __stat_dict = {}
            except TypeError as e:
                self.logger.info(f"Batter: {batter['PlayerID']}|{e}")
                __stat_dict = {}
            __stat_dict['PlayerID'] = batter['PlayerID']
            batter_stats.append(__stat_dict)

        pitcher_info = []
        pitcher_stats = []
        for pitcher in _json['pitcherResults']:
            pitcher_info.append({k: v for k, v in pitcher.items() if k != 'Stats'})
            try:
                __stat_dict = {k: v for k, v in pitcher['Stats']['pitching'].items()}
            except KeyError as e:
                self.logger.info(f"Pitcher: {pitcher['PlayerID']}|{e}")
                __stat_dict = {}
            except TypeError as e:
                self.logger.info(f"Pitcher: {pitcher['PlayerID']}|{e}")
                __stat_dict = {}

            __stat_dict['PlayerID'] = pitcher['PlayerID']
            pitcher_stats.append(__stat_dict)

        df_batter_info = pd.DataFrame(batter_info)
        df_pitcher_info = pd.DataFrame(pitcher_info)

        df_batter_stat = pd.DataFrame(batter_stats)
        df_pitcher_stat = pd.DataFrame(pitcher_stats)

        # concat batter and pitcher info and stats rows into one dataframe
        df_info = pd.concat([df_batter_info, df_pitcher_info], sort=False)
        df_stat = pd.concat([df_batter_stat, df_pitcher_stat], sort=False)
        # df_info = df_batter_info.concat([df_pitcher_info], sort=False)
        # df_stat = df_batter_stat.append(df_pitcher_stat, sort=False)

        return {
            'info': df_info,
            'stat': df_stat,
            'batter': {
                'info': df_batter_info,
                'stat': df_batter_stat
            },
            'pitcher': {
                'info': df_pitcher_info,
                'stat': df_pitcher_stat
            }
        }

    def player_details(self, player_id: str) -> pd.DataFrame:
        """
        Process an individual players statistics

        :param player_id: Ottoneu Player ID
        :return: DataFrame
        """
        headers = {
            'authority': 'ottoneu.fangraphs.com',
            'accept': '*/*',
            'origin': 'https://ottoneu.fangraphs.com',
            'x-requested-with': 'XMLHttpRequest',
            'content-player_type': 'application/x-www-form-urlencoded',
            'sec-fetch-site': 'same-origin',
            'sec-fetch-mode': 'cors',
            'referer': '{}/search'.format(self.ottoneu_base_url),
            'accept-encoding': 'gzip, deflate, br',
            'accept-language': 'en-US,en;q=0.9'}

        player_page = requests.get('{}/playercard?id={}'.format(self.ottoneu_base_url, player_id), headers=headers)
        soup = BeautifulSoup(player_page.text, 'html.parser')

        # milb_pitching_stats = process_player_page(soup, 'Pitching', 'MILB')
        # milb_batting_stats = process_player_page(soup, 'Batting', 'MILB')

        player_detail_stats = (self._process_player_page(soup, 'Batting', 'MLB')
                               .append(self._process_player_page(soup, 'Pitching', 'MLB')))
        player_detail_stats['player_id'] = player_id

        return player_detail_stats

    def league_transactions(self) -> pd.DataFrame:
        """
        Get Transaction Log of Transaction of Players in Fantasy League.

        :return: DataFrame
        """
        next_page = True
        df = pd.DataFrame()
        players = []
        teams = []

        while next_page:
            transaction_page = requests.get('{}{}'.format(self.ottoneu_base_url, 'transactions'))
            soup = BeautifulSoup(transaction_page.text, 'html.parser')
            next_page = soup.find_all('a', text="Next 50 transactions")
            if next_page:
                next_page = next_page[0].get('href', None)
            else:
                next_page = None
            table = soup.find_all('table')[0]
            for link in table.find_all('a'):
                url = link.get('href', None)
                if 'playercard' in url:
                    players.append(int(url[url.find('=') + 1:]))
                elif 'team' in url:
                    teams.append(int(url[url.find('=') + 1:]))

            df = df.append(pd.read_html(str(table))[0])
            self.logger.info(next_page)

        df['team_id'] = np.asarray(teams)
        df['player_id'] = np.asarray(players)
        df['Date'] = pd.to_datetime(df['Date'])
        df['Salary'] = pd.to_numeric(df['Salary'].str.replace('$', ''))
        return df

    def get_line_up(self) -> pd.DataFrame:
        """
        Get Lineup of Players in Fantasy League.

        :return:
        """



import datetime
import io
import logging
from datetime import timedelta

import numpy as np
import pandas as pd
import requests

from plyball.utils import sanitize_input, split_request


class StatCast(object):
    """
    StatCast Web Scraper
    """

    logger = logging.getLogger('statcast')
    c_handler = logging.StreamHandler()
    c_format = logging.Formatter('%(levelname)s: %(name)s - %(message)s')
    logger.addHandler(c_handler)

    urls = {
        'search': 'https://baseballsavant.mlb.com/statcast_search/csv?{}',
        'game': 'https://baseballsavant.mlb.com/statcast_search/csv?all=true&type=details&game_pk={}',
        'milb': 'https://www.mlb.com/prospects/stats/search/csv?{}'
    }

    default_parameters = {
        'all': 'true',
        'hfPT': 'FT',  # Pitch Type
        'hfAB': '',  # Plate Appearance
        'hfBBT': '',  # Batted Ball Type
        'hfPR': '',  # Pitch Result
        'hfZ': '',  # Gameday Zones
        'stadium': '',  # Stadium
        'hfBBL': '',  # Batted Ball Location
        'hfNewZones': '',  # Attack Zones
        'hfGT': 'R|',  # Season Type
        'hfC': '',  # Count
        'hfSea': '2019|',  # Season
        'hfSit': '',  # Situation
        'player_type': 'batter',  # PLayer Type
        'hfOuts': '',  # Outs
        'opponent': '',  # Opponent
        'pitcher_throws': '',  # Pitcher Handedness
        'batter_stands': '',  # Batter Handedness
        'hfSA': '',  # Quality of Contact
        'game_date_gt': '',
        'game_date_lt': '',
        'hfInfield': '',  # Inflield Alignment
        'team': '',
        'position': '',
        'hfOutfield': '',  # Outfield Alignment
        'hfRO': '',  # Runners On
        'home_road': '',
        'hfFlag': '',
        'hfPull': '',  # Batted Ball Direction
        'metric_1': '',
        'hfInn': '',
        'min_pitches': '0',
        'min_results': '0',
        'group_by': 'name',
        'sort_col': 'pitches',
        'player_event_sort': 'h_launch_speed',
        'sort_order': 'desc',
        'min_pas': 0,
        # 'type': 'details&'
    }

    @staticmethod
    def __postprocessing(data, team):
        """
        Scrub Data for Reply

        :param data: StatCast Data to Process
        :param team: Team to Select
        :return: DataFrame
        """
        data.replace(r'^\s*$', np.nan, regex=True, inplace=True)
        data.replace(r'^null$', np.nan, regex=True, inplace=True)

        valid_teams = ['MIN', 'PHI', 'BAL', 'NYY', 'LAD', 'OAK', 'SEA', 'TB', 'MIL', 'MIA',
                       'KC', 'TEX', 'CHC', 'ATL', 'COL', 'HOU', 'CIN', 'LAA', 'DET', 'TOR',
                       'PIT', 'NYM', 'CLE', 'CWS', 'STL', 'WSH', 'SF', 'SD', 'BOS', 'ARI', 'ANA', 'WAS']

        if team in valid_teams:
            data = data.loc[(data['home_team'] == team) | (data['away_team'] == team)]
        elif team is not None:
            raise ValueError('Error: invalid team abbreviation. Valid team names are: {}'.format(valid_teams))
        data = data.reset_index()
        return data

    def __single_game_request(self, game_pk):
        """
        Select Single game for Statcast
        :param game_pk:
        :return:
        """
        url = self.urls['game'].format(game_pk)
        s = requests.get(url, timeout=None).content
        data = pd.read_csv(io.StringIO(s.decode('utf-8')))
        return data

    def statcast_request(self, url: str, headers: bool = False, encoding: str = None):
        """
        Get Information from StatCast

        :param url: Url TO re
        :type url: str
        :param headers:
        :type headers: str
        :param encoding:
        :type encoding: str
        :return: DataFrame
        """

        self.logger.info(url)
        if headers:
            s = requests.get(url, timeout=None, headers=headers).content
        else:
            s = requests.get(url, timeout=None).content

        self.logger.info(s)
        if encoding:
            data = pd.read_csv(io.StringIO(s.decode(encoding)))
        else:
            data = pd.read_csv(io.StringIO(s))
        return data

    def large_request(self,
                      start_date: datetime.date,
                      end_date: datetime.date,
                      step: int,
                      parameters: dict,
                      url: str,
                      verbose: bool):
        """
        Break start and end date into smaller increments, collecting all data in small chunks and appending all
        results to a common dataframe.

        :param url:
        :param parameters:
        :param start_date:
        :param end_date:
        :param step:
        :param verbose:
        :return: pd.DataFrame
        """

        error_counter = 0  # count failed requests. If > X, break
        self.logger.info("This is a large query, it may take a moment to complete")
        dataframe_list = []
        intermediate_dt = start_date + timedelta(days=step)
        while intermediate_dt <= end_date:
            if (intermediate_dt.month < 4 and intermediate_dt.day < 15) or (
                    start_date.month > 10 and start_date.day > 14):
                if end_date.year > intermediate_dt.year:
                    self.logger.info('Skipping Off Season Dates')
                    start_date = start_date.replace(month=3, day=15, year=start_date.year + 1)
                    intermediate_dt = start_date + timedelta(days=step + 1)
                else:
                    break

            parameters['start_date'] = start_date.strftime('%Y-%m-%d')
            parameters['end_date'] = intermediate_dt.strftime('%Y-%m-%d')
            data = self.statcast_request(url.format(
                '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()])))
            if 'error' in data.columns:
                success = 0
                while success == 0:
                    parameters['start_date'] = start_date.strftime('%Y-%m-%d')
                    parameters['end_date'] = intermediate_dt.strftime('%Y-%m-%d')
                    data = self.statcast_request(url.format(
                        '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()])))
                    if data.shape[0] > 1:
                        dataframe_list.append(data)
                        success = 1
                    else:
                        error_counter += 1
                    if error_counter > 2:
                        intermediate_dt = intermediate_dt - timedelta(days=1)
                        parameters['start_date'] = start_date.strftime('%Y-%m-%d')
                        parameters['end_date'] = intermediate_dt.strftime('%Y-%m-%d')
                        smaller_data_1 = self.statcast_request(url.format(
                            '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()])))
                        if 'error' in data.columns:
                            self.logger.warning(
                                "Query unsuccessful for data from {} to {}. Skipping these dates.".format(
                                    start_date.strftime('%Y-%m-%d'),
                                    intermediate_dt.strftime('%Y-%m-%d')))
                        else:
                            dataframe_list.append(smaller_data_1)
                        error_counter = 0
                        break
            else:
                dataframe_list.append(data)
                if verbose:
                    self.logger.info(
                        "Completed sub-query from {} to {}".format(start_date.strftime('%Y-%m-%d'),
                                                                   intermediate_dt.strftime('%Y-%m-%d')))
            start_date = intermediate_dt + timedelta(days=1)
            intermediate_dt = intermediate_dt + timedelta(days=step + 1)
        if start_date < end_date:
            parameters['start_date'] = start_date.strftime('%Y-%m-%d')
            parameters['end_date'] = end_date.strftime('%Y-%m-%d')
            data = self.statcast_request(url.format(
                '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()])))
            dataframe_list.append(data)
            if verbose:
                logging.info(
                    "Completed sub-query from {} to {}".format(start_date.strftime('%Y-%m-%d'),
                                                               end_date.strftime('%Y-%m-%d')))

        final_data = pd.concat(dataframe_list, axis=0)
        return final_data

    def get_statcast_data(self,
                          start_dt: datetime.date = None,
                          end_dt: datetime.date = None,
                          team: str = None,
                          detail: bool = True,
                          verbose: bool = True) -> pd.DataFrame:
        """
        Pulls Statcast play-level data from Baseball Savant for a given date range. If no arguments are provided, this
        will return yesterday's statcast data. If one date is provided, it will return that date's Statcast data.

        :param start_dt: First date for which you want StatCast data
        :type start_dt: datetime.date
        :param end_dt: Last date for which you want StatCast data
        :type end_dt: datetime.date
        :param team: City abbreviation of the team you want data for (e.g. SEA or BOS)
        :type team: str
        :param detail: Provide Play-Level Detail
        :type detail: bool
        :param verbose:
        :type verbose: bool
        :return: DataFrame
        """

        parameters = self.default_parameters

        if detail:
            parameters['type'] = 'details&'

        small_query_threshold = 5
        data = self.large_request(start_dt,
                                  end_dt,
                                  step=small_query_threshold,
                                  parameters=parameters,
                                  url=self.urls['search'],
                                  verbose=verbose)

        data = self.__postprocessing(data, team)
        return data

    def single_game(self, game_pk: str, team: str = None) -> pd.DataFrame:
        """
        Pulls statcast play-level data from Baseball Savant for a single game, identified by its MLB game ID (game_pk
        in statcast data)

        :param self:
        :param game_pk: 6-digit integer MLB game ID to retrieve
        :type game_pk: str
        :param team:
        :type team: str
        :return: DataFrame
        """
        data = self.__single_game_request(game_pk)
        data = self.__postprocessing(data, team)
        return data

    @staticmethod
    def batter(start_dt: datetime.date, end_dt: datetime.date, player_id: str) -> pd.DataFrame:
        """
        Pulls statcast batter-level data from Baseball Savant for a given batter.

        :param start_dt: the first date for which you want data
        :type start_dt: datetime.date
        :param end_dt: the final date for which you want data
        :type end_dt: datetime.date
        :param player_id: the player's MLBAM ID. Find this by calling :func:plyball.player_id_lookup.
        :type player_id: str
        :return: DataFrame of StatCast Data for Batters
        :rtype: DataFrame
        """

        start_dt, end_dt, player_id = sanitize_input(start_dt, end_dt, player_id)
        url = ('https://baseballsavant.mlb.com/statcast_search/csv?all=true&hfPT=&hfAB=&hfBBT=&hfPR=&hfZ=&stadium'
               '=&hfBBL=&hfNewZones=&hfGT=R|PO|S|=&hfSea=&hfSit=&player_type=batter&hfOuts=&opponent'
               '=&pitcher_throws=&batter_stands=&hfSA=&game_date_gt={}&game_date_lt={}&batters_lookup%5B%5D={'
               '}&team=&position=&hfRO=&home_road=&hfFlag=&metric_1=&hfInn=&min_pitches=0&min_results=0&group_by'
               '=name&sort_col=pitches&player_event_sort=h_launch_speed&sort_order=desc&min_abs=0&type=details& ')
        df = split_request(start_dt, end_dt, player_id, url)
        return df

    @staticmethod
    def pitcher(start_dt: datetime.date, end_dt: datetime.date, player_id: str)-> pd.DataFrame:
        """
        Pulls statcast pitch-level data from Baseball Savant for a given pitcher.

        :param start_dt: the first_name date for which you want a player's sSatcast data
        :param end_dt: the final date for which you want data
        :param player_id: the player's MLBAM ID.
        :return: Dataframe of StatCast Data for Pitchers
        """

        start_dt, end_dt, player_id = sanitize_input(start_dt, end_dt, player_id)

        url = ('https://baseballsavant.mlb.com/statcast_search/csv?all=true&hfPT=&hfAB=&hfBBT=&hfPR=&hfZ=&stadium'
               '=&hfBBL=&hfNewZones=&hfGT=R|PO|S|=&hfSea=&hfSit=&player_type=pitcher&hfOuts=&opponent'
               '=&pitcher_throws=&batter_stands=&hfSA=&game_date_gt={}&game_date_lt={}&pitchers_lookup[]={'
               '}&team=&position=&hfRO=&home_road=&hfFlag=&metric_1=&hfInn=&min_pitches=0&min_results=0&group_by'
               '=name&sort_col=pitches&player_event_sort=h_launch_speed&sort_order=desc&min_abs=0&type=details& ')
        df = split_request(start_dt, end_dt, player_id, url)
        return df

    def milb_statcast(self,
                      start_date: datetime.date,
                      end_date: datetime.date,
                      details: bool,
                      verbose: bool,
                      **kwargs) -> pd.DataFrame:
        """
        Get StatCast Data for MILB players

        :param start_date:  the first date for which you want data

        :param end_date:  the last date for which you want data
        :param details: Aggregate Stats
        :param verbose: Increased Logging
        :return: Dataframe of MILB players
        :rtype: pd.DataFrame
        """
        small_query_threshold = 5
        parameters = {
            'pitcher_throws': '',
            'game_date_gt': '{}'.format(start_date.strftime('%Y-%m-%d')),
            'game_date_lt': '{}'.format(end_date.strftime('%Y-%m-%d')),
            'season': '',
            'home_away': '',
            'draft_year': '',
            'prospect': '',
            'player_type': 'batter',
            'sort_by': 'results',
            'sort_order': 'desc',
            'group_by': 'name',
            'min_results': '',
            'min_pa': ''
        }

        for k, v in kwargs.items():
            parameters[k] = v

        if details:
            parameters['is_aggregate'] = 'false'

        data = self.large_request(start_date,
                                  end_date,
                                  step=small_query_threshold,
                                  parameters=parameters,
                                  url=self.urls['milb'],
                                  verbose=verbose)

        return data



import logging
import re
import warnings
from typing import Literal

import numpy as np
import pandas as pd
import requests
import structlog
from bs4 import BeautifulSoup


class FanGraphs(object):
    """
    Class for scraping data from FanGraphs

    """
    _urls = {
        'leaders': 'https://www.fangraphs.com/api/leaders/major-league/data?{}',
        'milb_stats': 'https://www.fangraphs.com/api/leaders/minor-league/data?{}',
        'projections': 'https://www.fangraphs.com/api/projections?{}'
    }

    logger = structlog.get_logger("FanGraphs")

    def __init__(self):
        pass

    def __get_leaders_html(self,
                           player_type: Literal['bat', 'pit'],
                           start_season: int,
                           end_season: int = None,
                           league: int = 'all',
                           qualified: int = 'y',
                           split_season: int = 1,
                           **kwargs):
        """
        Get Leaderboard HTML form Fangraphs to convert into a DataFrame

        :param player_type: pit or bat
        :type player_type: Literal['bat','pit']
        :param start_season:
        :param end_season:
        :param league:
        :param qualified: The minimum pla
        :param split_season:
        :return: BeautifulSoup
        """

        if player_type == 'pit':
            columns = ['c'] + [str(num) for num in range(3, 322)]
        elif player_type == 'bat':
            columns = ['c'] + [str(num) for num in range(3, 305)]
        else:
            columns = ['c']

        parameters = {
            'pos': 'all',
            'stats': player_type,
            'lg': league,
            'qual': qualified,
            'type': ','.join(columns),
            'season': start_season,
            'month': 0,
            'season1': end_season,
            'ind': split_season,
            'team': ','.join(kwargs.get('team', [''])),
            'rost': kwargs.get('rost', ''),
            'age': kwargs.get('age', ''),
            'filter': kwargs.get('filter', ''),
            'players': kwargs.get('players', '0'),
            'page': '1_999999999',
            'pageitems': '2000000000'
        }

        s = requests.get(
            self._urls['leaders'].format('&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()
        self.logger.info(self._urls['leaders'].format('&'.join(['{}={}'.format(k, v) for k, v in parameters.items()])))
        return s

    def __get_leader_table(self,
                           player_type: Literal['pit', 'bat'],
                           start_season: int,
                           **kwargs) -> pd.DataFrame:

        data = self.__get_leaders_html(player_type, start_season, **kwargs)['data']

        __data = pd.DataFrame(data=data)

        self.logger.info("Loaded DataFrame with {} rows".format(__data.shape[0]),
                         shape=__data.shape,
                         example=__data.head(1))

        # replace empty strings with NaN
        __data.replace(r'^\s*$', np.nan, regex=True, inplace=True)

        # convert percentage strings to floats FB% duplicated
        percentages = [column for column in __data.columns if '%' in column]
        for col in percentages:
            if not __data[col].empty:
                if pd.api.types.is_string_dtype(__data[col]):
                    __data[col] = __data[col].str.strip(' %')
                    __data[col] = __data[col].str.strip('%')
                    __data[col] = pd.to_numeric(__data[col], errors='coerce') / 100.

        __data.columns = [
            re.sub(r"\W", '_',
                   column.replace('%', '_percent').replace('-', '_negative_').replace('+', '_positive_').replace(
                       '__',
                       '_'))
            for column in __data.columns]

        with warnings.catch_warnings():
            """
            Pandas will throw a performance warning due to fragmentation of the data. This is expected behavior and
            can be ignored. The DataFrame is copied to avoid this warning from being thrown for the user.
            """
            warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)
            __data['player_type'] = player_type

        return __data.copy()

    def get_pitching_table(self,
                           start_season: int,
                           end_season: int = None,
                           league: str = 'all',
                           minimum_innings_pitched: int = None,
                           split_season: bool = True):
        """
        Get individual pitching data from LeaderBoard.

        :param start_season:
        :param end_season:
        :param league:
        :param minimum_innings_pitched:
        :param split_season:
        :return:
        """
        if start_season is None:
            raise ValueError(
                "You need to provide at least one season to collect data for. Try pitching_leaders(season) or "
                "pitching_leaders(start_season, end_season).")
        if end_season is None:
            end_season = start_season
        return self.__get_leader_table(player_type='pit',
                                       start_season=start_season,
                                       end_season=end_season,
                                       league=league,
                                       qualified='0' if minimum_innings_pitched is None else minimum_innings_pitched,
                                       ind=1 if split_season else 0)

    def get_batting_table(self,
                          start_season,
                          end_season: int = None,
                          league: str = 'all',
                          minimum_plate_appearances: int = None,
                          split_season: bool = True):
        """
        Get Individual Batting Stats from the LeaderBoard.

        :param start_season:
        :param end_season:
        :param league:
        :param minimum_plate_appearances:
        :param split_season:
        :return:
        """
        if start_season is None:
            raise ValueError(
                "You need to provide at least one season to collect data for. Try pitching_leaders(season) or "
                "pitching_leaders(start_season, end_season).")
        if end_season is None:
            end_season = start_season
        return self.__get_leader_table(player_type='bat',
                                       start_season=start_season,
                                       end_season=end_season,
                                       league=league,
                                       qualified='0' if minimum_plate_appearances is None else minimum_plate_appearances,
                                       ind=1 if split_season else 0)

    def get_team_batting_table(self,
                               start_season: int,
                               end_season: int = None,
                               league: str = 'all',
                               minimum_plate_appearances: int = None,
                               split_season: bool = True):
        """
        Get Team Batting Stats from the LeaderBoard.

        :param start_season: First Season to return data
        :param end_season: Last Season to return data
        :param league: American (AL) or National (NL) Baseball League
        :param minimum_plate_appearances: Minimum number of Plate Appearances (Not supplied only minimum_innings_pitched players will be returned)
        :param split_season: Return Season Splits for players
        :return:
        """
        if start_season is None:
            raise ValueError(
                "You need to provide at least one season to collect data for. "
                "Try pitching_leaders(season) or pitching_leaders(start_season, end_season).")
        if end_season is None:
            end_season = start_season
        return self.__get_leader_table(player_type='bat',
                                       start_season=start_season,
                                       end_season=end_season,
                                       league=league,
                                       qualified='y' if minimum_plate_appearances is None else minimum_plate_appearances,
                                       ind=1 if split_season else 0,
                                       team=['0', 'to'])

    def get_team_pitch_table(self,
                             start_season: int,
                             end_season: int = None,
                             league: str = 'all',
                             minimum_innings_pitched: int = None,
                             split_season: bool = True):
        """
        Get Team Pitching Stats from teh LeaderBoard.


        :param start_season: First Season to return data
        :param end_season: Last Season to return data
        :param league: American (AL) or National (NL) Baseball League
        :param minimum_innings_pitched: Minimum number of Pitching Appearances (Not supplied only minimum_plate_appearances players will be returned)
        :param split_season: Return Season Splits for players

        :return: DataFrame
        """
        if start_season is None:
            raise ValueError(
                "You need to provide at least one season to collect data for. "
                "Try pitching_leaders(season) or pitching_leaders(start_season, end_season).")
        if end_season is None:
            end_season = start_season
        return self.__get_leader_table(player_type='pit',
                                       start_season=start_season,
                                       end_season=end_season,
                                       league=league,
                                       qualified='y' if minimum_innings_pitched is None else minimum_innings_pitched,
                                       ind=1 if split_season else 0,
                                       team=['0', 'to'])

    def get_milb_pitching_table(self,
                                start_season: int,
                                end_season: int = None,
                                league: str = 'all',
                                minimum_innings_pitched: int = None,
                                split_season: int = 0):
        """
        Get Minor League Pitching Stats from the Prospects sections

        :param start_season: First Season to return data
        :param end_season: Last Season to return data
        :param league: American (AL) or National (NL) Baseball League
        :param minimum_innings_pitched: Minimum number of Plate Appearances (Not supplied only minimum_innings_pitched players will be returned)
        :param split_season: Return Season Splits for players
        :return: DataFrame
        """
        parameters = {
            'pos': 'all',
            'level': 0,
            'stats': 'pit',
            'lg': league,
            'qual': '0' if minimum_innings_pitched is None else minimum_innings_pitched,
            'type': '0',
            'team': '',
            'season': start_season,
            'seasonEnd': start_season if end_season is None else end_season,
            'ind': split_season,
            'org': '',
            'splitTeam': 'false'
        }
        # pos = all & level = 0 & lg = 2 & stats = bat & qual = y & type = 0 & team = & season = 2023 & seasonEnd = 2023 & org = & ind = 0 & splitTeam = false

        json = requests.get(
            url=self._urls['milb_stats'].format(
                '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()

        df = pd.DataFrame(json)
        df['Name'] = df.Name.apply(lambda x: BeautifulSoup(x, 'lxml').a.contents[0])

        return df

    def get_milb_batting_table(self,
                               start_season: int,
                               end_season: int = None,
                               league: str = 'all',
                               minimum_plate_appearances: int = None,
                               split_season: bool = True) -> pd.DataFrame:
        """
        Get Minor League Pitching from the Prospects sections.

        :param start_season: First Season to return data
        :param end_season: Last Season to return data
        :param league: American (AL) or National (NL) Baseball League
        :param minimum_plate_appearances: Minimum number of Plate Appearances (Not supplied only minimum_innings_pitched players will be returned)
        :param split_season: Return Season Splits for players
        :return: DataFrame
        """
        parameters = {
            'pos': 'all',
            'level': 0,
            'lg': league,
            'stats': 'bat',
            'qual': '0' if minimum_plate_appearances is None else minimum_plate_appearances,
            'team': '',
            'type': '0',
            'season': start_season,
            'seasonEnd': start_season if end_season is None else end_season,
            'org': '',
            'ind': 1 if split_season else 0,
            'splitTeam': 'false'
        }

        # ?pos = all & level = 0 & lg = all & stats = pit & qual = 0 & type = 0 & team = & season = 2023 & seasonEnd = 2023 & org = & ind = 0 & splitTeam = false & players = & sort = 25, 1

        json = requests.get(
            url=self._urls['milb_stats'].format(
                '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()

        self.logger.info("JSON", json=json)
        df = pd.DataFrame(json)
        # df['Name'] = df.Name.apply(lambda x: BeautifulSoup(x, 'lxml').a.contents[0])

        return df

    def get_zip_bat_projections(self) -> pd.DataFrame:
        parameters = {
            'type': 'steamer',
            'stats': 'bat',
            'pos': 'all',
            'team': '0',
            'players': '0',
            'lg': 'all',
        }

        json = requests.get(
            url=self._urls['projections'].format(
                '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()
        df = pd.DataFrame(json)

        return df

    def get_zip_pitching_projections(self) -> pd.DataFrame:
        parameters = {
            'type': 'steamer',
            'stats': 'pit',
            'pos': 'all',
            'team': '0',
            'players': '0',
            'lg': 'all',
        }

        json = requests.get(
            url=self._urls['projections'].format(
                '&'.join(['{}={}'.format(k, v) for k, v in parameters.items()]))).json()
        df = pd.DataFrame(json)

        return df

    def get_daily_game_log(
            self,
            player_id: int,
            position: str,
            season: int = None,
            start_date: str = None,
            end_date: str = None,
            **kwargs):
        """
        https://www.fangraphs.com/api/players/game-log?playerid=10155&position=OF&type=0&season=2023
        :return:
        """

        parameters = {
            'playerid': player_id,
            'position': position,
            'type': 0,
            'season': season,
            'start': start_date,
            'end': end_date
        }

        url = 'https://www.fangraphs.com/api/players/game-log?{}'.format(
            '&'.join(['{}={}'.format(k, v) for k, v in parameters.items() if v != '' and v is not None]))

        logging.info(url)

        json = requests.get(url).json()

        df = pd.DataFrame(json['mlb'])

        return df



import pandas as pd

from plyball.player_map import get_player_map
from plyball.fangraphs import FanGraphs
from plyball.ottoneu import Ottoneu


def calculate_sabr_points(fangraph_stats, position_type):
    """
    Function for calculating number of points in SABR fantasy leagues.
    TODO: make function for more generic leagues

    :param fangraph_stats:
    :param position_type:
    :return:
    """
    scoring = {
        'batting': {'AB': -1.0,
                    'H': 5.6,
                    '2B': 2.9,
                    '3B': 5.7,
                    'HR': 9.4,
                    'BB': 3.0,
                    'HBP': 3.0,
                    'SB': 1.9,
                    'CS': -2.8},
        'pitching': {'IP': 5.0,
                     'SO': 2.0,
                     'BB': -3.0,
                     'HBP': -3.0,
                     'HR': -13.0,
                     'SV': 5.0,
                     'HLD': 4.0,
                     }
    }

    fangraph_stats['fantasy_points'] = 0
    if position_type == 'pitching':
        fangraph_stats['IP'] = pd.to_numeric(fangraph_stats['IP']).round(0) \
                               + (pd.to_numeric(fangraph_stats['IP'])
                                  - pd.to_numeric(fangraph_stats['IP']).round(0)[1:]) \
                               * 3.33
        for stat, points in scoring['pitching'].items():
            fangraph_stats['fantasy_points'] = (fangraph_stats['fantasy_points'] +
                                                (pd.to_numeric(fangraph_stats[stat], errors='coerce') * points))

    elif position_type == 'batting':
        for stat, points in scoring['batting'].items():
            fangraph_stats['fantasy_points'] = (fangraph_stats['fantasy_points'] +
                                                (pd.to_numeric(fangraph_stats[stat], errors='coerce') * points))

    return fangraph_stats['fantasy_points']


def player_yearly_summary_data():
    """

    :return:
    """
    __player_map = get_player_map()
    __ottoneu = Ottoneu(186).players()['info']
    __ottoneu['PlayerID'] = pd.to_numeric(__ottoneu['PlayerID'], errors='coerce')
    __ottoneu = __ottoneu.merge(__player_map,
                                right_on='OTTONEUID',
                                left_on='PlayerID')

    __fangraph_pitching = FanGraphs().get_pitching_table(2019)
    __fangraph_pitching['fantasy_points'] = calculate_sabr_points(__fangraph_pitching, 'pitch')
    __fangraph_batting = FanGraphs().get_batting_table(2019)
    __fangraph_batting['fantasy_points'] = calculate_sabr_points(__fangraph_batting, 'batting')

    __fangraph = pd.merge(__fangraph_pitching,
                          __fangraph_batting,
                          on='player_id',
                          how='outer', suffixes=('_pitch', '_bat'))

    df = __ottoneu.merge(__fangraph, right_on='player_id', left_on='IDFANGRAPHS', how='left')
    df.Points = df.Points.fillna(df.fantasy_points_bat.fillna(0) + df.fantasy_points_pitch.fillna(0))

    return df
